
from numpy import *
import matplotlib.pyplot as pp


def plot_all(mes, th, ic1, ic2, ss = None, nf = 64, sample_f = 1.0,
             logss = False, sqrtmes = False, plotf = None):
    '''Plots nxn graphics, with confidence intervals and threshold.
       If ss == True, plots ss in the diagonal.
       Already expects data in power form: abs(x)^2'''
    if logss:
        ss = log(ss)

    if sqrtmes:
        mes = sqrt(mes)
        th = sqrt(th)
        ic1 = sqrt(ic1)
        ic2 = sqrt(ic2)

    x = sample_f*arange(nf)/(2.0*nf)
    n = mes.shape[0]
    pp.figure(figsize=(8, 6), dpi=300)

    for i in range(n):
        for j in range(n):
            pp.subplot(n,n,i*n+j+1)
            #over = mes[i,j][mes[i,j]>th[i,j]]
            #overx = x[mes[i,j]>th[i,j]]
            over = mes[i,j]
            overx = x

            #Old code
            #under = mes[i,j][mes[i,j]<=th[i,j]]
            #underx = x[mes[i,j]<=th[i,j]]
            #pp.plot(x, th[i,j], 'r:', x, ic1[i,j], 'k:', x, ic2[i,j], 'k:',
            #        overx, over, 'b-', underx, under, 'r-')
            if i!=j:
                pp.plot(x, th[i,j], 'k--', x, ic1[i,j], 'k:', x, ic2[i,j], 'k:',
                        overx, over, 'r-')
                #Complicated code for underthreshold painting
                k = 0
                while(k < nf):
                    while(mes[i,j,k] >= th[i,j,k]):
                        k = k+1
                        if (k == nf): break
                    if (k == nf): break
                    kold = k
                    while(mes[i,j,k] < th[i,j,k]):
                        k = k+1
                        if (k == nf): break
                    pp.plot(x[kold:k], mes[i,j,kold:k], 'g-')


            pp.ylim(-0.05,1.05)
            if (i < n-1):
                pp.xticks([])
            if (j > 0):
                pp.yticks([])

            if plotf != None:
                pp.xlim([0, plotf])

        if (ss is not None):
            ax = pp.subplot(n,n,i*n+i+1).twinx()
            ax.plot(sample_f*arange(nf)/(2.0*nf), ss[i,i,:], color='b')
            if logss:
                ax.set_ylim(ymin = ss[i,i,:].min(), ymax = ss[i,i,:].max())
            else:
                ax.set_ylim(ymin = 0, ymax = ss[i,i,:].max())

            if (i < n-1):
                ax.set_xticks([])

            if plotf != None:
                ax.set_xlim(xmin = 0, xmax = plotf)

    pp.show(block=False)

def pdc_plot(pdc, ss = None, nf = 64, sample_f = 1.0):
    '''Plots nxn graphics.
       If ss == True, plots ss in the diagonal.
       Expects data in complex form. Does: abs(x)^2 before plotting.'''
    n = pdc.shape[0]
    pdc = pdc*pdc.conj()
    pp.figure(figsize=(8, 6), dpi=300)

    for i in range(n):
        for j in range(n):
            pp.subplot(n,n,i*n+j+1)
            pp.plot(sample_f*arange(nf)/(2.0*nf), pdc[i,j,:])
            pp.ylim(-0.05,1.05)
            if (i < n-1):
                pp.xticks([])
            if (j > 0):
                pp.yticks([])
        if (ss != None):
            ax = pp.subplot(n,n,i*n+i+1).twinx()
            ax.plot(sample_f*arange(nf)/(2.0*nf), ss[i,i,:], color='g')
            ax.set_ylim(ymin = 0, ymax = ss[i,i,:].max())
            if (i < n-1):
                ax.set_xticks([])
    pp.show(block=False)



# -*- coding:utf-8 -*-
"""
Created on 06/11/2009

@author: Carlos Stein
"""

from numpy import *
from numpy.random import randn
from numpy.random import multivariate_normal as mnorm

import time

# from scipy import weavee
# from scipy.weave import converters

y = array([[1936, 1.0, 0.9,  40],
           [1937, 0.8, 0.8, 115],
           [1938, 0.8, 0.8, 100],
           [1939, 1.4, 1.3,  80],
           [1940, 1.2, 1.4,  60],
           [1941, 1.0, 1.2,  40],
           [1942, 1.5, 1.7,  23],
           [1943, 1.9, 1.8,  10],
           [1944, 1.5, 1.6,  10],
           [1945, 1.5, 1.5,  25],
           [1946, 1.5, 1.5,  75],
           [1947, 1.6, 2.0, 145],
           [1948, 1.8, 2.5, 130],
           [1949, 2.8, 2.7, 130],
           [1950, 2.5, 2.9,  80],
           [1951, 2.5, 2.5,  65],
           [1952, 2.4, 3.1,  20],
           [1953, 2.1, 2.4,  10],
           [1954, 1.9, 2.2,   5],
           [1955, 2.4, 2.9,  10],
           [1956, 2.4, 2.5,  60],
           [1957, 2.6, 2.6, 190],
           [1958, 2.6, 3.2, 180],
           [1959, 4.4, 3.8, 175],
           [1960, 4.2, 4.2, 120],
           [1961, 3.8, 3.9,  50],
           [1962, 3.4, 3.7,  35],
           [1963, 3.6, 3.3,  20],
           [1964, 4.1, 3.7,  10],
           [1965, 3.7, 3.9,  15],
           [1966, 4.2, 4.1,  30],
           [1967, 4.1, 3.8,  60],
           [1968, 4.1, 4.7, 105],
           [1969, 4.0, 4.4, 105],
           [1970, 5.2, 4.8, 105],
           [1971, 5.3, 4.8,  80],
           [1972, 5.3, 4.8,  65]])
sun = y[:, [3, 2]].transpose()

''' Some pre-specified models for ready use.
    ar_models(2) is the sunspot-melanoma data.
    lam is the free parameter is models that have one.'''


def ar_models(id, lam=0.0):

    models = [
        # 0
        [array([[[0.2, 0], [0, 0], [0.3, -0.2]],
                [[0, 0], [0.8, -0.1], [0.4, -0.1]],
                [[0, 0], [-0.1, 0.2], [0.4, 0.1]]], dtype=float),
         identity(3)],
        # 1
        [array([[[4, -4], [3, 3]], [[0, -2], [2, -3]]], dtype=float).reshape(2, 2, 2) / 20,
            array([[0.7, 0], [0, 2]], dtype=float)],
        # 2 sunspot melanoma
        sun,
        # 3
        [array([[[4, 3, -2], [-2, -5, 3]], [[4, -2, 1], [-4, 0, 3]]], dtype=float).reshape(2, 2, 3) / 20,
            array([[0.7, 0.3], [0.3, 2]], dtype=float)],
        # 4 JAS Daniel (12)
        [array([[[0.2, 0], [-0.4, -0.2], [0.3, 0]],
                [[lam, 0], [0.8, -0.1], [0.4, 0.0]],
                [[0, 0.5], [-0.1, 0.2], [0.4, 0.1]]], dtype=float),
         identity(3)],

    ]
    return models[id]

# def ar_data_R(A, er = None, m = 1000):
# #   from rpy2.robjects import r as r_
# #   import rpy2.rinterface as ri_
# #   import rpy2.robjects as ro_
#
#    '''Simulate ar-model from A matrix
#
#      Input:
#        A(n, n, p) - AR model (n - number of signals, p - model order)
#        er(n) - variance of innovations
#        m - length of simulated time-series
#
#      Output:
#        data(n, m) - simulated time-series
#    '''
#    if er == None:
#        er = ones(A.shape[0])
#
#    ri_.initr()
#    r_('library(dse1)')
#
#    n = A.shape[0]
#    A = concatenate((eye(n).reshape(n,n,1), -A), axis = 2)
#    ri_.globalEnv["A"] = ri_.FloatSexpVector(A.ravel())
#    ri_.globalEnv["dim"] = ri_.IntSexpVector(A.shape[::-1])
#    ri_.globalEnv["er"] = ri_.FloatSexpVector(er)
#    ro_.globalEnv["m"] = m
#    ro_.globalEnv["n"] = n
#    return array(r_('simulate(ARMA(A = array(A, dim), B = diag(1,n)), sd = sqrt(er), sampleT = m)$output')).T

# def ar_data_old(A, er=None, m=1000, dummy=100):


def ar_data(A, er=None, m=1000, dummy=100):
    '''Simulate ar-model from A matrix

      Input:
        A(n, n, p) - AR model (n - number of signals, p - model order)
        er(n) - variance of innovations
        m - length of simulated time-series

      Output:
        data(n, m) - simulated time-series
    '''

    if (A.ndim == 2):
        A.resize(A.shape[0], A.shape[1], 1)

    n = A.shape[0]
    p = A.shape[2]
    if er.any() == None:
        er = identity(n)
    if er.ndim == 1:
        er = diag(er)

    print(time.clock())
    w = mnorm(zeros(n), er, m + dummy - p)
    print(time.clock())
    data = zeros([n, m + dummy])
    for i in arange(p, m + dummy):
        for j in arange(p):
            data[:, i] = data[:, i] + dot(A[:, :, j], data[:, i - j - 1])
        data[:, i] = data[:, i] + w[i - p]
    print(time.clock())

    return data[:, dummy:]

# def ar_data(A, er = None, m = 1000, dummy = 100):
#     '''Simulate ar-model from A matrix
#
#       Input:
#         A(n, n, p) - AR model (n - number of signals, p - model order)
#         er(n) - variance of innovations
#         m - length of simulated time-series
#
#       Output:
#         data(n, m) - simulated time-series
#     '''
#
#     if (A.ndim == 2):
#         A.resize(A.shape[0], A.shape[1], 1)
#
#     n = A.shape[0]
#     p = A.shape[2]
#     if er is None:
#         er = identity(n)
#     if er.ndim == 1:
#         er = diag(er)
#
#     w = mnorm(zeros(n), er, m+dummy-p)
#     data = zeros([n, m+dummy])
#     code = '''
#         for (int i = p; i < m+dummy; i++) {
#             for (int j = 0; j < p; j++) {
#                 for (int k = 0; k < n; k++) {
#                     double s = 0;
#                     for (int t = 0; t < n; t++) {
#                         data[t,i];
#                         s += A(k,t,j)*data(t,i-j-1);
#                     }
#                     data(k,i) = data(k,i) + s;
#                 }
#             }
#             for (int k = 0; k < n; k++) {
#                 data(k,i) += w(i-p,k);
#             }
#         }
#     '''
#     weave.inline(code, ['data', 'p', 'm', 'dummy', 'n', 'A', 'w'],
#                        type_converters=converters.blitz, compiler = 'gcc')
#
#
#     return data[:,dummy:]



# -*- coding:utf-8 -*-
"""
Created on 06/11/2009

@author: Carlos Stein
"""

from numpy import *
import scipy.signal as sig
import time

from pypdc.analysis import pdc_alg, pdc_and_plot
from pypdc.plotting import pdc_plot
from matplotlib.pyplot import gcf

def aPDC(data, p, step = 50, se = 100, preproc = False):
    '''data(m,n,nd) -> data, m = #trials, n = #channels, nd = #time samples
       p -> VAR model order
       se -> efective sample memory of adaptative model'''

    m,n,nd = data.shape

    if preproc:
        data = pre_proc_ding_99(data)

    A, er = AMVAR(data, p, se)

    for i in arange(p, nd, step):
        pd = pdc_alg(A[i], er[i], metric = 'diag')
        pdc_plot(pd)
        canvas = gcf().canvas
        canvas.start_event_loop(timeout=0.2)

        #time.sleep(1)

    return A, er

def AMVAR (data, p, se = 100):
    '''data(m,n,nd) -> data, m = #trials, n = #channels, nd = #samples'''

    m,n,nd = data.shape

    A = zeros([nd,n,n*p])
    er = zeros([nd,n,n])
    era = zeros([nd,n,n])
    C = mat(identity(n*p))

    #Stein modification: (usar cf em torno de 0.02 para um AR(2,2,2))
    cf = 2.0*m/float64(se+m)
    print(cf)

    for i in arange(p,nd):
        C = (1.0/(1.0-cf))*C
        if i == p:
            Wt = mat(data[:,:,i-1::-1].transpose(0,2,1).reshape(m,-1))
        else:
            Wt = mat(data[:,:,i-1:i-p-1:-1].transpose(0,2,1).reshape(m,-1))

        for j in arange(m):
            C = C*(identity(n*p) - Wt[j].T*Wt[j]*C/(Wt[j]*C*Wt[j].T + 1.0))

        K = Wt*C
        #Yn = data(:,:,i)
        Z = data[:,:,i] - Wt*A[i-1].T
        A[i] = A[i-1] + Z.T*K

        if cf > 0:
            er[i] = (1-cf)*er[i-1] + (cf/float64(m))*Z.T*Z
            era[i] = er[i]/(1.0-(1.0-cf)**i)
        else:
            er[i] = (n/(n+1.0))*er[i-1] + (1.0/(m*(n+1.0)))*Z.T*Z
            era[i] = er[i]

        #if (i%(nd/10) == 0 and i*m > 200):
        #    print 'nd', i
        #    print 'C', C
        #    #print 'Z', Z
        #    print 'dA', Z.T*K


    er = era
    A = A.reshape(nd,n,p,n).transpose(0,1,3,2)

    return A, er


def pre_proc_ding_99(data):

    m,n,nd = data.shape

    #normalize per trial
    data = sig.detrend(data)
    data = data/std(data, axis = 2).reshape(m,n,1)

    #normalize per time
    if m > 1:
        data = data - mean(data, 0).reshape(1,n,nd)
        data = data/std(data, axis = 0).reshape(1,n,nd)

    return data



# -*- coding:utf-8 -*-
"""
Created on 22/10/2009

@author: Carlos Stein
"""

from numpy import *
import time
from pypdc.ar_fit import nstrand

from pypdc.ar_data import ar_data

def bootstrap(method_func, nd, nm, A, er, 
              nf, alpha = 0.05, metric = None, thres = True):
    '''Bootstrap statistics'''
    
    n = A.shape[0]
    mes = empty([nm, n, n, nf])
    th  = zeros([nm, n, n, nf])
    #ic1 = empty([nm, n, n, nf])
    #ic2 = empty([nm, n, n, nf])
    maxp = A.shape[2]
    tbegin = time.clock()
    for i in range(nm):
        if (i%1 == 0):
            print('nm:', i, 'time:', time.clock()-tbegin)
        #Generate data from AR
        data = ar_data(A, er, nd)
        #Estimate AR parameters with Nuttall-Strand
        Aest, erest = nstrand(data, maxp = maxp)
        #Calculate the connectivity and statistics
        if (metric == None):
            mes[i] = abs(method_func(Aest, erest, nf = nf))**2
        else:
            mes[i] = abs(method_func(Aest, erest, nf = nf, metric = metric))**2
            
    if thres:
        for a in range(n):
            for b in range(n):
                
                if a == b:
                    continue
                
                Anew = A.copy()
                Anew[a,b,:] = 0
                
                for i in range(nm):
                    if (i%1 == 0):
                        print('nm:', i, 'time:', time.clock()-tbegin)
                    #Generate data from AR
                    data = ar_data(Anew, er, nd)
                    #Estimate AR parameters with Nuttall-Strand
                    Aest, erest = nstrand(data, maxp = maxp)
                    #Calculate the connectivity and statistics
                    if (metric == None):
                        aux = abs(method_func(Aest, erest, nf = nf))**2
                        th[i][a][b] = aux[a,b,:]
                    else:
                        aux = abs(method_func(Aest, erest, nf = nf, metric = metric))**2
                        th[i][a][b] = aux[a,b,:]
        

    so = sort(mes, axis = 0)
    ic1 = so[(alpha/2)*nm]
    ic2 = so[(1-alpha/2)*nm]
    #bvar = var(mes, axis = 0)

    if (metric == None):
        mes0 = abs(method_func(A, er, nf = nf))**2
    else:
        mes0 = abs(method_func(A, er, nf = nf, metric = metric))**2
    
    if thres:
        to = sort(th, axis = 0)
        th = to[(1-alpha)*nm]
    else:
        th = ones(ic1.shape)*0#0.5/sqrt(nd) #TODO: pensar melhor como fazer.
    
    return mes0, th, ic1, ic2



# -*- coding:utf-8 -*-


from numpy import array,zeros,eye,kron,dot,exp,pi,cos,sin,diag,ones,tril,resize,finfo
from numpy.core.fromnumeric import reshape
from numpy.core.umath import isnan
from numpy.lib.scimath import log, sqrt
from numpy.linalg import pinv, inv, eig
from scipy.linalg.basic import det
from scipy.linalg import schur
# from scipy.linalg import solve_continuous_lyapunov as solve_lyapunov
from scipy.linalg import solve_sylvester

# scipy.linalg.solve_sylvester(a, b, q)[source]
# Computes a solution (X) to the Sylvester equation (AX + XB = Q).

eps = finfo(float).eps.item()

def nstrand(u, maxp = 30, simplep = True):
    '''
    %   Calculate the coeficients of multi-channel auto-regressive matrix using
    %   Nuttall-Strand algorithm (a generalization of single channel harmonic
    %                             method)
    %
    %   Input parameters:
    %     IP     - Ordem of autoregressive model (integer)
    %     u      - Complex matrix with NUMCHS channels of sample data
    %
    %   Output parameters:
    %     PF     - Covariance matrix of NUMCHS x NUMCHS of linear forward
    %              prediction error
    %     A      - Complex array of forward linear prediction matrix
    %              coefficients
    %     PB     - Complex backward linear prediction error covariance array
    %     B      - Complex array of backward linear prediction matrix
    %              coefficients
    '''
    [lx,cx]=u.shape
    if lx > cx:
        lx, cx = cx, lx  # Swapping the number of channels and record length
        u = u.T
        print ('Input matrix is probably transposed.')
        return
    NUMCHS=lx      # Number of channels.
    MAXORDER=200   # Maximum order of AR model allowed for calculation.
    N=max(u.shape) # N - Number of samples per channel.
    IP = maxp

    #    Initialization
    ISTAT=0
    if (IP > MAXORDER):
        ISTAT=3
        print('IP > 200')
        return

    ef=array(u)                    # Eq. (15.91)
    eb=array(u)                    # Eq. (15.91)
    pf=dot(u, u.transpose())       # Eq. (15.90)
    pb=array(pf)                   # Eq. (15.90)
    M=0
    #    Main Loop
    while 1:
        # Update estimated covariance errors  Eq. (15.89)
        pfhat=dot(ef[:,M+1:N],ef[:,M+1:N].transpose())
        pbhat=dot(eb[:,M:N-1],eb[:,M:N-1].transpose())
        pfbhat=dot(ef[:,M+1:N],eb[:,M:N-1].transpose())
        M=M+1
        # Calculate estimated partial correlation matrix - Eq. (15.98)
        #              (Nuttall-Strand algorithm only)
        RHO=solve_sylvester(dot(pfhat,inv(pf)),dot(inv(pb),pbhat),2*pfbhat)
        #RHO=solve_lyapunov(dot(pfhat,inv(pf)),dot(inv(pb),pbhat),-2*pfbhat)
        #  Update forward and backward reflection coeficients
        #% Eqs. (15.73),(15.74),(15.78) (Nuttall-Strand algorithm)
        AM=dot(-RHO,inv(pb))
        BM=dot(-RHO.transpose(),inv(pf))
        dimA=AM.shape[0]
        dimB=BM.shape[0]
        if M == 1:
            A=zeros((1,dimA,dimA),float)
            B=zeros((1,dimB,dimB),float)
        else:
            A=resize(A,(M,dimA,dimA))
            B=resize(B,(M,dimB,dimB))
        A[M-1,:,:]=AM
        B[M-1,:,:]=BM
        #  Update forward and backward covariance error  - Eqs. (15.75),(15.76)
        pf=pf-dot(dot(AM,BM),pf)
        pb=pb-dot(dot(BM,AM),pb)

        #  Update forward and backward predictor coeficients - Eqs.(15.84),(15.85)
        if not (M == 1):
            for K in range(1,M):
                temp1=A[K-1,:,:].copy()
                A[K-1,:,:]=A[K-1,:,:]+dot(AM,B[M-K-1,:,:])
                B[M-K-1,:,:]=B[M-K-1,:,:]+dot(BM,temp1)

        Tef=array(ef)
        ef[0:NUMCHS,range(N-1,M-1,-1)]=ef[:,range(N-1,M-1,-1)]+dot(AM,eb[:,range(N-2,M-2,-1)])
        eb[0:NUMCHS,range(N-1,M-1,-1)]=eb[:,range(N-2,M-2,-1)]+dot(BM,Tef[:,range(N-1,M-1,-1)])

        #  Verify if model order is adequate
        if M == IP:
            A=-A
            B=-B
            break

    if (simplep):
        return A.transpose(1,2,0), abs(pf)/N
    else:
        return pf,A,pb,B,ef,eb,ISTAT

def mvar(u, MaxIP = 30, alg=1, criterion=1, return_ef = False):
    '''
    %
    %[IP,pf,A,pb,B,ef,eb,vaic,Vaicv] = mvar(u,MaxIP,alg,criterion,return_ef)
    %
    % input: u     - data rows
    %        MaxIP - externally defined maximum IP (default = 30)
    %        alg   - for algorithm (1: Nutall-Strand; 2: mlsm; 3: Vieira-Morf; 
    %                               4: QR artfit)
    %        criteria for order choice - 0: MDL (not implemented);
    %                                     1: AIC; 2: Hanna-Quinn; 3: BIC-Schwarz;
    %                                     4: FPE; 5: fixed order in MaxIP
    %                                     6(not yet): estimate up to max order
    %                                     Negative(not yet) - keep criterion changes
    % output: 
    %
    %% Description
    %
    %
    %
    %
    %% References
    %
    % [1] Lutkepohl H (2005). New Introduction to Multiple Time Series Analysis. 
    %                         Springer-Verlag. 
    '''

    # StopFlag = False
    [nSegLength,nChannels] = u.transpose().shape

    # if criterion < 0:
    #     stopFlag = True
    #     criterion=abs(criterion)

    if criterion == 5:    # Fixed order at IP
        IP = maxIP
        if alg == 1:
            # print('u: ',u, 'MaxIP: ', MaxIP)
            npf, na, npb, nb, nef, neb, ISTAT = nstrand(u,MaxIP,False)
            pf =  npf/nSegLength
            A = na.transpose(1,2,0)
        else:
            print('Algorithm not implemented yet in Python.')

        vaic = max(u.shape)*log(det(pf)) + 2*nChannels*nChannels*IP # Akaike's 
        Vaicv =  vaic
        # if (not return_ef):
        return IP, pf, A, npb, nb, nef,neb,vaic,Vaicv
        # else:
        #     return na.transpose(1,2,0), nef

    vaicv=0
    if MaxIP == 0:
        MaxOrder = 30
        UpperboundOrder = round(3*sqrt(nSegLength)/nChannels)
        #% Marple Jr. page 409
        #% Suggested by Nuttall, 1976.
        UpperboundOrder = min([MaxOrder, UpperboundOrder])
    else:
        MaxOrder=MaxIP
        UpperboundOrder=MaxIP

    IP=1
    Vaicv=zeros((MaxOrder+1,1), float)

    while IP <= UpperboundOrder:
        if alg == 1:
            npf, na, npb, nb, nef, neb, ISTAT = nstrand(u,IP,False)
        else:
            print('ALG TO BE IMPLEMENTED: \n 2: Least squares estimator; \n 3: Vieira-Morf algorithm; \n 4: ARfit estimator.')

        if criterion == 1:   # Akaike's Information Criterion (AIC)  # (4.3.2)(Lutkepöhl '85)
            vaic = max(u.shape)*log(det(npf)) + 2*nChannels*nChannels*IP 

        elif criterion == 2: # Hanna-Quin (HQ)                       # (4.3.8)(Lutkepöhl '85)
            vaic = max(u.shape) * log(det(npf)) + 2*log(log(length(max(u.shape))))*nChannels*nChannels*IP

        elif criterion == 3: # Schwartz (1978) (SC or BIC criterion) # (4.3.9)(Lutkepöhl '85)
            vaic = max(u.shape) * log(det(npf)) + log(max(u.shape))*nChannels*nChannels*IP

        elif criterion == 4: # Final prediction error (Akaike, 1970) # (4.3.1)(Lutkepöhl '85)
            vaic = log(det(npf) * ((max(u.shape) + nChannels*IP+1)/(max(u.shape)-nChannels*IP-1))**nChannels)
        else:
            print('CRITERION TO BE IMPLEMENTED: ')

        Vaicv[IP,0] = vaic
        print('IP=', IP, '; vaic=', vaic)

        if (vaic > vaicv) and not (IP == 1): # Akaike condition
            vaic = vaicv
            print('vaic > vaicv ==> break')
            break
            # if not StopFlag:
            #     break

        vaicv = vaic
        pf = array(npf)
        A  = array(na)
        ef = array(nef)
        if alg == 1 or alg == 3:  # alg = Nutall-Strand and Vieira-Morf
            B  = array(nb)
            eb = array(neb)
            pb = array(npb)
        else:       # review status for backward prediction in clmsm
            B = []  # empty
            eb = [] # 
            pb = [] #

        IP=IP+1

    IP = IP - 1
    vaic = vaicv
    Vaicv = Vaicv[range(1,IP+1),0]
    Vaicv.shape = (Vaicv.size,1)

    if alg == 1:
        pf =  pf/nSegLength
        A = A.transpose(1,2,0)
    else:
        print('Algorithm not implemented yet in Python.')

    if (not return_ef):
        return IP, pf, A, npb, nb, nef, neb, vaic, Vaicv
    else:
        return IP, pf, A, npb, nb, nef, neb, vaic, Vaicv



from numpy import *
import matplotlib.pyplot as plt


def plot_all(mes, th, ic1, ic2, ss=True, nf=64, sample_f=1.0,
             logss=True, sqrtmes=False, plotf=None):
    '''Plots nxn graphics, with confidence intervals and threshold.
       If ss == True, plots ss in the diagonal.
       Already expects data in power form: abs(x)^2'''
    if logss:
        ss = log(ss)

    if sqrtmes:
        mes = sqrt(mes)
        th = sqrt(th)
        ic1 = sqrt(ic1)
        ic2 = sqrt(ic2)

    x = sample_f*arange(nf)/(2.0*nf)
    n = mes.shape[0]
    plt.figure()

    for i in range(n):
        for j in range(n):
            plt.subplot(n, n, i*n+j+1)
            #over = mes[i,j][mes[i,j]>th[i,j]]
            #overx = x[mes[i,j]>th[i,j]]
            over = mes[i, j]
            overx = x
            if i != j:
                plt.plot(x, th[i, j], 'k--', overx, over, 'r-')  # b  --> r

                # plt.plot(x, th[i, j], 'k--', x, ic1[i, j], 'k:', x, ic2[i, j], 'k:',
                #          overx, over, 'r-')  # b  --> r

            # Complicated code for underthreshold painting
            k = 0
            while(k < nf):
                while(mes[i, j, k] >= th[i, j, k]):
                    k = k+1
                    if (k == nf):
                        break
                if (k == nf):
                    break
                kold = k
                while(mes[i, j, k] < th[i, j, k]):
                    k = k+1
                    if (k == nf):
                        break

                if i != j:
                    plt.plot(x[kold:k], mes[i, j, kold:k], 'g-')  # r -> b

            plt.ylim(-0.05, 1.05)
            if (i < n-1):
                plt.xticks([])
            if (j > 0):
                plt.yticks([])

            if plotf != None:
                plt.xlim([0, plotf])

        if (ss is not None):
            ax = plt.subplot(n, n, i*n+i+1).twinx()
            ax.plot(sample_f*arange(nf)/(2.0*nf), ss[i, i, :], color='b')
            if logss:
                ax.set_ylim(ymin=ss[i, i, :].min(), ymax=ss[i, i, :].max())
            else:
                ax.set_ylim(ymin=0, ymax=ss[i, i, :].max())

            if (i < n-1):
                ax.set_xticks([])

            if plotf != None:
                ax.set_xlim(xmin=0, xmax=plotf)

    plt.show(block=False)


def pdc_plot(pdc2, ss=None, nf=64, sample_f=1.0):
    '''Plots nxn graphics.
       If ss == True, plots ss in the diagonal.
       Expects data in complex form. Does: abs(x)^2 before plotting.'''
    n = pdc2.shape[0]
    # pdc = pdc*pdc.conj() # pdc2 is already a squared pdc estimate
    pdc2 = abs(pdc2)
    plt.figure()

    for i in range(n):
        for j in range(n):
            plt.subplot(n, n, i*n+j+1)
            plt.plot(sample_f*arange(nf)/(2.0*nf), pdc2[i, j, :])
            plt.ylim(-0.05, 1.05)
            if (i < n-1):
                plt.xticks([])
            if (j > 0):
                plt.yticks([])
        if (ss != None):
            ax = plt.subplot(n, n, i*n+i+1).twinx()
            ax.plot(sample_f*arange(nf)/(2.0*nf), ss[i, i, :], color='g')
            ax.set_ylim(ymin=0, ymax=ss[i, i, :].max())
            if (i < n-1):
                ax.set_xticks([])
    plt.show(block=False)



# -*- coding:utf-8 -*-

import numpy as np
import matplotlib.pyplot as pp
from scipy.linalg import inv
import scipy.signal as sig
from scipy.stats import f

import cProfile

from pypdc.ar_data import ar_data
from pypdc.ar_fit import mvar
import pypdc.asymp as as_
from pypdc.plotting import *
import pypdc.plotting as pl_
import pypdc.bootstrap as bt_

def logo():
    print(' _____       _____  _____   _____ ')
    print('|  __ \     |  __ \|  __ \ / ____|')
    print('| |__) |   _| |__) | |  | | |     ')
    print('|  ___/ | | |  ___/| |  | | |     ')
    print('| |   | |_| | |    | |__| | |____ ')
    print('|_|    \__, |_|    |_____/ \_____|')
    print('        __/ |                     ')
    print('       |___/                      ')

def list_to_array(data):
    '''Converts a list to an array'''
    d = data[0].reshape(1,-1)
    for i in range(1,len(data)):
        d = concatenate([d, data[i].reshape(1,-1)], axis = 0)
    return d

def pre_data(data, normalize = True, detrend = True):
    if (detrend):
        data = sig.detrend(data)

    if (normalize):
        data = data/std(data, axis = 1).reshape(-1,1)

    return data

def A_to_f(A, nf = 64):
    '''Calculates A(f), in the frequency domain

    Input:
        A(n, n, r) - recurrence matrix (n - number of signals, r - model order)
        nf - frequency resolution

    Output:
        AL(nf, n, n)
    '''

    n, n, r = A.shape

    # Exponents contains the array of the fft exponents, with all frequencies for all lags
    exponents = (-1j*pi*kron(arange(nf),(arange(r)+1.0))/nf).reshape(nf,r)
    # Af performs the multiplications of exp(ar) by the matrix A, for all frequencies
    # as funcoes repeat e transpose sao truques para possibilitar o calculo vetorial
    Af = (A.reshape(n,n,1,r).repeat(nf, axis=2)*exp(exponents)).transpose([2,0,1,3])
    # fft sums the value for all lags
    AL = eye(n) - sum(Af, axis = 3)

    return AL


def pc_alg(A, e_cov, nf = 64):
    '''Calculates the Partial Coherence
        A -> autoregressive matrix
        e_cov -> residues
        nf -> number of frequencies
        '''
    n, n, r = A.shape

    e_cov = mat(e_cov)
    AL = A_to_f(A, nf)
    pc = empty(AL.shape, dtype = 'complex')
    for i in range(nf):
        ALi = mat(AL[i])
        ps = ALi.T*e_cov.I*ALi.conj()
        d = ps.diagonal()
        m = kron(d,d).reshape(n,n)
        pc[i] = ps/sqrt(m)
    return pc.transpose(1,2,0)

def ss_alg(A, e_cov, nf = 64):
    '''Calculates the Spectral density (SS)
        A -> autoregressive matrix
        e_cov -> residues
        nf -> number of frequencies
        '''
    n, n, r = A.shape

    AL = A_to_f(A, nf)
    ss = empty(AL.shape, dtype = 'complex')
    for i in range(nf):
        H = mat(AL[i]).I
        ss[i] = H*e_cov*H.T.conj()

#     print(ss[5])
    return ss.transpose(1,2,0)

def ss_coh_alg(A, e_cov, nf = 64):
    '''Calculates the Spectral density (SS) and Coherence (coh)
        A -> autoregressive matrix
        e_cov -> residues
        nf -> number of frequencies
        '''
    n, n, r = A.shape

    AL = A_to_f(A, nf)
    coh = empty(AL.shape, dtype = 'complex')
    ss = empty(AL.shape, dtype = 'complex')
    for i in range(nf):
        H = mat(AL[i]).I
        ss[i] = H*e_cov*H.T.conj()
        d = ss[i].diagonal()
        m = kron(d,d).reshape(n,n)
        coh[i] = ss[i]/sqrt(m)
    return ss.transpose(1,2,0), coh.transpose(1,2,0)

def coh_alg(A, e_cov, nf = 64):
    '''Calculates the Coherence (coh)
        A -> autoregressive matrix
        e_cov -> residues
        nf -> number of frequencies
        '''
    n, n, r = A.shape

    AL = A_to_f(A, nf)
    coh = empty(AL.shape, dtype = 'complex')
    for i in range(nf):
        H = mat(AL[i]).I
        ss = H*e_cov*H.T.conj()
        d = ss.diagonal()
        m = kron(d,d).reshape(n,n)
        coh[i] = ss/sqrt(m)
    return coh.transpose(1,2,0)

def pdc_alg(A, e_cov, nf = 64, metric = 'gen'):
    '''Generates spectral general (estatis. norm) PDC matrix from AR matrix

      Input:
        A(n, n, r) - recurrence matrix (n - number of signals, r - model order)
        e_cov(n, n) - error covariance matrix
        nf - frequency resolution

      Output:
        PDC(n, n, nf) - PDC matrix
    '''

    n, n, r = A.shape
    if metric == 'euc':
        nornum = ones(n)
        norden = identity(n)
    elif metric == 'diag':
        nornum = 1/diag(e_cov)
        norden = diag(1/diag(e_cov))
    else: #metric == 'gen'
        nornum = 1/diag(e_cov)
        norden = inv(e_cov)
    print('A====: ', A)
    AL = A_to_f(A, nf)

    ALT = AL.transpose([0,2,1])
    #dPDC = sum(dot(ALT,norden)*ALT.conj(), axis = -1).reshape(nf,-1)
    dPDC = sum(dot(ALT,norden)*ALT.conj(), axis = -1).reshape(nf,-1)
    nPDC = AL*sqrt(nornum).reshape(-1,1)
    PDC = nPDC/sqrt(abs(dPDC)).reshape(nf,1,n).repeat(n, axis = 1)
    return PDC.transpose(1,2,0)

def dtf_alg(A, er, nf = 64):
    '''Generates spectral not normalized DTF matrix from AR matrix

      Input:
        A(n, n, r) - recurrence matrix (n - number of signals, r - model order)
        e_cov(n, n) - error covariance matrix
        nf - frequency resolution

      Output:
        DTF(n, n, nf) - PDC matrix
    '''

    n, n, r = A.shape
    #nor = ones(n) # dtf_one nao tem normalizacao

    AL = A_to_f(A, nf)
    HL = empty(AL.shape, dtype=complex)
    for i in range(nf):
        HL[i] = inv(AL[i])

    # normalization by sum(ai ai* sig)
    dDTF = sum(HL*HL.conj(), axis = 2)
    nDTF = HL
    DTF = nDTF/sqrt(abs(dDTF)).reshape(nf,n,1).repeat(n, axis = 2)

    return DTF.transpose(1,2,0)


def pdc_ss_coh(data, maxp = 30, nf = 64, detrend = True):
    '''Interface that returns the PDC, SS and coh'''

    if(type(data) == 'list'):
        data = list_to_array(data)

    if (detrend):
        data = sig.detrend(data)

    A, er = mvar(data, maxp)
    return abs(pdc_alg(A, er, nf))**2, abs(ss_alg(A, er, nf))**2, abs(coh_alg(A, er, nf))**2


def pdc(data, maxp = 30, nf = 64, detrend = True, normalize = False,
        fixp = False, ss = True, metric = 'diag'):
    '''Generates spectral PDC matrix from data array

      Input:
        data(n, m) - data matrix (n - number of signals, m - data length)
        maxp - maximum order for estimated AR model
        nf - frequency resolution
        detrend - Shall the data be detrended
        SS - Shall calculate the SS also
        metric - which PDC to use ('euc', 'diag' or 'gen')

      Output:
        PDC(n, n, nf) - PDC matrix
        ss(n, n, nf) - Parametric cross spectral matrix
    '''

    if(type(data) == type([])):
        data = list_to_array(data)


    data = pre_data(data, normalize, detrend)

    crit = 0 #AIC
    if fixp:
        crit = 1

    A, er = mvar(data, maxp, criterion=crit)

    print('data:', data.shape)
    print('A:', A.shape)

    #print A

    if (ss):
        return pdc_alg(A, er, nf, metric = metric), ss_alg(A, er, nf)
    else:
        return pdc_alg(A, er, nf, metric = metric)


def coh(data, maxp = 30, nf = 64, detrend = True, normalize = False, fixp = False, ss = True):
    '''Interface that calculate the Coherence from data'''

    if(type(data) == type([])):
        data = list_to_array(data)


    data = pre_data(data, normalize, detrend)

    crit = 0 #AIC
    if fixp:
        crit = 1

    A, er = mvar(data, maxp, criterion=crit)

    if (ss):
        return coh_alg(A, er, nf), ss_alg(A, er, nf)
    else:
        return coh_alg(A, er, nf)


def dtf(data, maxp = 30, nf = 64, detrend = True, normalize = False, fixp = False, ss = True):
    '''Interface that calculate the Coherence from data'''

    if(type(data) == type([])):
        data = list_to_array(data)


    data = pre_data(data, normalize, detrend)

    crit = 0 #AIC
    if fixp:
        crit = 1

    A, er = mvar(data, maxp, criterion=crit)


    if (ss):
        return dtf_alg(A, er, nf), ss_alg(A, er, nf)
    else:
        return dtf_alg(A, er, nf)

def ss(data, maxp = 30, nf = 64, detrend = True, normalize = False, fixp = False, ss = True):
    '''Interface that calculate the Coherence from data'''

    if(type(data) == type([])):
        data = list_to_array(data)


    data = pre_data(data, normalize, detrend)

    crit = 0 #AIC
    if fixp:
        crit = 1

    A, er = mvar(data, maxp, criterion=crit)

    return ss_alg(A,er,nf)

def pc(data, maxp = 30, nf = 64, detrend = True, normalize = False, fixp = False, ss = True):
    '''Interface that calculate the Coherence from data'''

    if(type(data) == type([])):
        data = list_to_array(data)


    data = pre_data(data, normalize, detrend)

    crit = 0 #AIC
    if fixp:
        crit = 1

    A, er = mvar(data, maxp, criterion=crit)

    if (ss):
        return pc_alg(A, er, nf), ss_alg(A, er, nf)
    else:
        return pc_alg(A, er, nf)


#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
#%
#%  Computes granger causality index
#%
#%  Input:
#%    D(n, N) - data (n channels)
#%    MaxIP - externaly defined maximum IP
#%
#%  Output:
#%    Gr(n, n) - Granger causalit index
#%
#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
#
#function [Gr] = alg_ganger(u, maxIP)
#
#[n N] = size(u);
#
#[IP,pf,A,pb,B,ef,eb,vaic,Vaicv] = mvar(u,maxIP,[0 0]);
#
#va = diag(pf);
#
#va_n = zeros(n, n);
#
#for iu = 1:n
#  aux_u = u;
#  aux_u(iu,:) = [];
#  [IP,pf,A,pb,B,ef,eb,vaic,Vaicv] = mvar(aux_u,maxIP,[0 0]);
#  aux = diag(pf)';
#  va_n(iu,:) = cat(2, aux(1:iu-1), 0, aux(iu:n-1));
#end
#
#Gr = zeros(n, n);
#for iu = 1:n
#  for ju = 1:n
#    if (iu == ju) continue; end
#    Gr(iu,ju) = log(va_n(ju,iu)/va(iu));
#  end
#end

def gci(data, maxp = 30, detrend = True):

    n = data.shape[0]

    if (detrend):
        data = sig.detrend(data)

    A0, er0 = mvar(data, maxp)
    va0 = diag(er0)

    gci = zeros([n,n])
    for i in arange(n):
        aux_data = delete(data, i, 0)
        A1, er1 = mvar(aux_data, maxp)
        va1 = diag(er1)
        va1 = insert(va1, i, 0)
        gci[:,i] = log(float64(va1)/va0)

    return gci


def gct(data, maxp = 30, detrend = True):
    '''Asymptotic statistics for Wald statistic of the GC in time
        data -> data
        maxp -> max mvar order
    '''

    if (detrend):
        data = sig.detrend(data)

    A, e_var = mvar(data, maxp)

    return as_.asymp_gct(data, A, e_var)

def igct(data, maxp = 30, detrend = True):
    '''Asymptotic statistics for Wald statistic of instantaneous GC
        x -> data
        maxp -> max mvar order
        alpha -> confidence margin
    '''

    if (detrend):
        data = sig.detrend(data)

    A, e_var = mvar(data, maxp)

    n, nd = data.shape

    return as_.asymp_igct(e_var, nd)

def white_test(data, maxp = 30, h = 20):

    A, res = mvar(data, maxp, return_ef=True)

    n,n,p = A.shape

    return as_.asymp_white(data, res, p, h)


#def gct(data, maxp = 30, detrend = True):
#    #TODO: esta errado, apagar.
#    n,T = data.shape
#
#    if (detrend):
#        data = sig.detrend(data)
#
#    A0, er0 = mvar(data, maxp)
#    va0 = diag(er0)
#
#    p = A0.shape[2] #TODO: p pode variar depois. fixar para A1?
#    print p
#
#    gci = zeros([n,n])
#    for i in arange(n):
#        aux_data = delete(data, i, 0)
#        A1, er1 = mvar(aux_data, maxp)
#        va1 = float64(diag(er1))
#        va1 = insert(va1, i, 0)
#        gci[:,i] = ((va1-va0)/(n*p))/(va0/(T-n*p-1))
#
#    gct = f.cdf(gci, n*p, T-n*p-1)
#
#    return gct

#### TODO need to introduce mvar model order selection criteria and algorithm in the arguments


#=================================================================
#Output the graphics of PDC results with confidence intervals and
#threshold curves at a significance level of aplha = 0.01

def pdc_full(data, maxp = 20, nf = 64, sample_f = 1,
                   ss = True, alpha = 0.01, metric = 'info',
                   detrend = True, normalize = False,
                   stat = 'asymp', n_boot = 1000, fixp = False,
                   plotf = None):
    '''Calculate PDC and asymptotic statistics from data, and plot results.'''
    
    logo()
    
    print('alpha_full=', alpha)
    print('maxIP_full=', maxp)

    if(type(data) == type([])):
        data = list_to_array(data)

    data = pre_data(data, normalize, detrend)
    n,nd = data.shape
    #Estimate AR parameters with Nuttall-Strand
    alg = 1
    crit = 1 #AIC

    if fixp:
        crit = 5

    print('crit_full=',crit)
    print('alg_full=',alg)

    print('data.shape: ', data.shape)
    #def mvar(u, MaxIP = 30, alg=1, criterion=1, return_ef = False):
    IP, pf, Aest, pb, B, ef, eb, vaic, Vaicv = mvar(data, maxp, alg, crit, return_ef=True)

    erest = pf
    stat='asymp'

    if stat == 'asymp':
        #def asymp_pdc(x, A, nf, e_var, p, metric='info', alpha=0.01):
        mes, ss, coh, th, ic1, ic2, patdf, patden = as_.asymp_pdc(data, Aest, nf, erest,
                                                    IP, metric = metric, alpha = alpha)
    elif stat == 'boot':
        mes, th, ic1, ic2 = bt_.bootstrap(pdc_alg, nd, n_boot, Aest, erest,
                            nf, alpha = alpha, metric = metric)
    else:
        mes = pdc_alg(Aest, erest, nf, metric)
        th = zeros(mes.shape)
        ic1 = zeros(mes.shape)
        ic2 = zeros(mes.shape)

    ssm = ss_alg(Aest, erest, nf)
    
    plot_all(mes, th, ic1, ic2, nf = nf,
              ss = ssm, sample_f = sample_f, plotf = plotf)

    
    
#=========================================================
#Output the weighted connectivity matrix (|PDC|^2 matrix)
#with significant connectivities at alpha = 0.01

    
def pdc_pdc2_th(data, maxp = 20, nf = 64, sample_f = 1,
                   ss = True, alpha = 0.01, metric = 'info',
                   detrend = True, normalize = False,
                   stat = 'asymp', n_boot = 1000, fixp = False,
                   plotf = None):
    '''Calculate PDC and asymptotic statistics from data'''

    if(type(data) == type([])):
        data = list_to_array(data)

    data = pre_data(data, normalize, detrend)
    n,nd = data.shape
    #Estimate AR parameters with Nuttall-Strand
    alg = 1
    crit = 1 #AIC

    if fixp:
        crit = 5

    #def mvar(u, MaxIP = 30, alg=1, criterion=1, return_ef = False):
    IP,pf,Aest,pb,B,ef,eb,vaic,Vaicv = mvar(data, maxp, alg, crit, return_ef=True)

    erest = pf
    stat='asymp'

    if stat == 'asymp':
        #def asymp_pdc(x, A, nf, e_var, p, metric='info', alpha=0.01):
        mes, ss, coh, th, ic1, ic2, patdf, patden = as_.asymp_pdc(data, Aest, nf, erest,
                                                     IP, metric = metric, alpha = alpha)
    elif stat == 'boot':
        mes, th, ic1, ic2 = bt_.bootstrap(pdc_alg, nd, n_boot, Aest, erest,
                            nf, alpha = alpha, metric = metric)
    else:
        mes = pdc_alg(Aest, erest, nf, metric)
        th = zeros(mes.shape)
        ic1 = zeros(mes.shape)
        ic2 = zeros(mes.shape)

    ssm = ss_alg(Aest, erest, nf)
    return mes, th


#M and Th are n x n matrices:
def compMatrices(M, Th):
    for i in  range(0, len(M)):
        for j in  range(0, len(M)):
             if M[i,j] <= Th[i,j]:
                    M[i,j] = 0
             else:
                 pass
                
def mat_c(A):
    L = []
    for i in range(len(A)):
        for j in range(len(A[i])):
             L.append(A[i,j].max())
    
    M = np.asmatrix(L)
    M = M.reshape(len(A), len(A))
    return M
            

def pdc_matrix(data, maxp = 20, nf = 64, sample_f = 1,
                   ss = True, alpha = 0.01, metric = 'info',
                   detrend = True, normalize = False,
                   stat = 'asymp', n_boot = 1000, fixp = False,
                   plotf = None):
        logo()
        
        pdc2_th = pdc_pdc2_th(data, maxp = 20, nf = 64, sample_f = 1,
                   ss = True, alpha = 0.01, metric = 'info',
                   detrend = True, normalize = False,
                   stat = 'asymp', n_boot = 1000, fixp = False,
                   plotf = None)
        
        M = mat_c(pdc2_th[0])
        Th = mat_c(pdc2_th[1])
        compMatrices(M, Th)
        
        print("\n PDC connectivity matrix:")
        return M
        

#=================================================================
            
    

def coh_full(data, maxp = 5, nf = 64, sample_f = 1,
             ss = True, alpha = 0.05, detrend = True, normalize = False, stat = 'asymp', n_boot = 1000, fixp = False, metric = None):
    measure_full(data, 'coh', maxp, nf, sample_f, ss, alpha, detrend, normalize, stat = stat, n_boot = n_boot, fixp = fixp)

def dtf_full(data, maxp = 5, nf = 64, sample_f = 1,
             ss = True, alpha = 0.05, detrend = True, normalize = False, stat = 'asymp', n_boot = 1000, fixp = False, metric = None):
    measure_full(data, 'dtf2', maxp, nf, sample_f, ss, alpha, detrend, normalize, stat = stat, n_boot = n_boot, fixp = fixp)

def ss_full(data, maxp = 5, nf = 64, sample_f = 1,
             ss = True, alpha = 0.05, detrend = True, normalize = False, stat = 'asymp', n_boot = 1000, fixp = False, metric = None):
    measure_full(data, 'ss', maxp, nf, sample_f, ss, alpha, detrend, normalize, stat = stat, n_boot = n_boot, fixp = fixp)

def pc_full(data, maxp = 5, nf = 64, sample_f = 1,
             ss = True, alpha = 0.05, detrend = True, normalize = False, stat = 'asymp', n_boot = 1000, fixp = False, metric = None):
    measure_full(data, 'pc', maxp, nf, sample_f, ss, alpha, detrend, normalize, stat = stat, n_boot = n_boot, fixp = fixp)

def measure_full(data, measure, maxp = 5, nf = 64, sample_f = 1,
                 ss = True, alpha = 0.05, detrend = True,
                 normalize = False, stat = 'asymp', n_boot = 1000, fixp = False):
    '''Interface that calculates measure from data, calculates asymptotics statistics and plots everything.
       measure: 'dtf2', 'coh', 'ss', 'pc'
       '''


    if(type(data) == type([])):
        data = list_to_array(data)

    n,nd = data.shape

    data = pre_data(data, normalize, detrend)

    crit = 0 #AIC
    if fixp:
        crit = 1
    algo = 0
    #Estimate AR parameters with Nuttall-Strand
    Aest, erest = mvar(data, maxp, alg=algo,criterion=crit)

    print( 'A:', Aest)
    #erest = (erest+erest.T)/2   #TODO: conferir isso.
    print('evar:', erest)
    #Calculate the connectivity and statistics

    if stat == 'asymp':
        if (measure == 'dtf2'):
            mes, th, ic1, ic2, patdf, patden = as_.asymp_dtf(data, Aest, nf, erest,
                                              maxp, alpha = alpha)
        if (measure == 'coh'):
            mes, th, ic1, ic2 = as_.asymp_coh(data, Aest, nf, erest,
                                              maxp, alpha = alpha)
        if (measure == 'ss'):
            mes, th, ic1, ic2 = as_.asymp_ss(data, Aest, nf, erest,
                                              maxp, alpha = alpha)
        if (measure == 'pc'):
            mes, th, ic1, ic2 = as_.asymp_pc(data, Aest, nf, erest,
                                              maxp, alpha = alpha)
    elif stat == 'boot':
        methcall = globals()[measure + '_alg']
        mes, th, ic1, ic2 = bt_.bootstrap(methcall, nd, n_boot, Aest, erest,
                                          nf, alpha = alpha)
    else:
        methcall = var()[measure + '_alg']
        mes = methcall(Aest, erest, nf)
        th = zeros(mes.shape)
        ic1 = zeros(mes.shape)
        ic2 = zeros(mes.shape)

    if (ss == True):
        ssm = ss_alg(Aest, erest, nf)
    else:
        ssm = None

    plot_all(mes, th, ic1, ic2, nf = nf, ss = ssm, sample_f = sample_f)


def pdc_and_plot(data, maxp = 30, nf = 64, sample_f = 1, ss = True, metric = 'gen',
                 detrend = True, normalize = False, fixp = False):
    '''Interface that calculates PDC from data and plots it'''
    if(type(data) == type([])):
        data = list_to_array(data)

    pdc_, ss_ = pdc(data, maxp, nf, detrend = detrend, normalize = normalize,
                    fixp = fixp, metric = metric)
    if(not ss):
        ss_ = None

    pdc_plot(pdc_, ss_, nf, sample_f)

def coh_and_plot(data, maxp = 30, nf = 64, sample_f = 1, ss = True, metric = None,
                 detrend = True, normalize = False, fixp = False):
    measure_and_plot(data, 'coh', maxp, nf, sample_f, ss, fixp = fixp)

def dtf_and_plot(data, maxp = 30, nf = 64, sample_f = 1, ss = True, metric = None,
                 detrend = True, normalize = False, fixp = False):
    measure_and_plot(data, 'dtf2', maxp, nf, sample_f, ss, fixp = fixp)

def ss_and_plot(data, maxp = 30, nf = 64, sample_f = 1, ss = True, metric = None,
                detrend = True, normalize = False, fixp = False):
    measure_and_plot(data, 'ss', maxp, nf, sample_f, ss, fixp = fixp)

def pc_and_plot(data, maxp = 30, nf = 64, sample_f = 1, ss = True, metric = None,
                detrend = True, normalize = False, fixp = False):
    measure_and_plot(data, 'pc', maxp, nf, sample_f, ss, fixp = fixp)

def measure_and_plot(data, measure, maxp = 30, nf = 64, sample_f = 1, ss = True,
                     detrend = True, normalize = False, fixp = False):
    '''Interface that calculates PDC from data and plots it'''
    if(type(data) == type([])):
        data = list_to_array(data)


    if (measure == 'dtf'):
        alg = dtf
    if (measure == 'coh'):
        alg = coh
    if (measure == 'ss'):
        alg = ss
    if (measure == 'pc'):
        alg = pc

    if (ss):
        mea, ss_ = alg(data, maxp, nf, detrend = detrend, normalize = normalize,
                    fixp = fixp, ss = True)
    else:
        mea = alg(data, maxp, nf, detrend = detrend, normalize = normalize,
                    fixp = fixp, ss = False)
        ss_ = None

    pdc_plot(mea, ss_, nf, sample_f)



"""
PyPDC
=========

PDC is a Python package for calculating asymptotic Partial Directed Coherence estimations for brain connectivity analysis.

"""

name = "pypdc"

#Current version:
__version__ = "0.0.6"







# -*- coding:utf-8 -*-


# In this file we calculate the asymptotic statistics for all measures, including
# first and second order asymptotic aproximations.

from numpy import *
import scipy.stats as st
#from scipy.stats import cov as cov
from numpy import cov as cov
from numpy.linalg import cholesky
from numpy.linalg import eigh
from numpy.linalg import inv
from numpy.linalg import pinv

from scipy.sparse.linalg import inv as sinv

from scipy.sparse import coo_matrix
from scipy.sparse import csr_matrix
from scipy.sparse import lil_matrix
from scipy.sparse import dia_matrix
from scipy import sparse 
from matplotlib.pyplot import xcorr
from pypdc.ar_fit import *
from numpy.linalg import eig
from numpy.linalg.linalg import LinAlgError 
import sys
import numpy as np
from numba import jit

from timeit import default_timer as timer

# These functions are used to make code more readable.

#def vec(x): return np.array(x.ravel('F')).T
def vec(x): return csr_matrix.transpose(csr_matrix(x.ravel('F')))

def vec3(x): return csr_matrix(x.ravel('F'))

def vec2(x): return np.transpose(x.ravel('F'))
# def vec2(x): return (np.ravel(x, order='F')).T


def O(n): return np.zeros([n, n], dtype=float)


def I(n): return np.array(identity(n, dtype=float))


def cat(a, b, ax): return concatenate((a, b), axis=ax)


def mdiag(a): return np.array(diag(diag(a)))


# def diagtom(a): return diagonal(csr_matrix(a.reshape(-1)))
# def diagtom(a): return csr_matrix(np.diag(np.array(a).reshape(-1)))
def diagtom(a): 
    m,m =  a.shape
    return csr_matrix(dia_matrix((a, 0), shape = (m, m)))


def diagtom2(a):
    c,m = a.shape
    print("***> diagtom c ", c, " m ", m)
    data = np.diag(a)
    ind = [(m+1)*k for k in np.arange(m)]
    b = lil_matrix((m*m, m*m), dtype=np.float64)
    for k in np.arange(m):
        index = (m+1)*k
        b[index,index] = a[k,k]
    return b.tocsr()


def A_to_f(A, nf=64):
    '''Calculates A(f), in the frequency domain.

    Input:
        A(n, n, r) - recurrence matrix (n - number of signals, r - model order)
        nf - frequency resolution

    Output:
        AL(nf, n, n)
    '''

    n, n, r = A.shape
    A = np.array(A)

    # Exponents contains the array of the fft exponents, with all frequencies for all lags
    expoents = (-1j * pi * np.kron(np.arange(nf), (np.arange(r) + 1.0)) / nf).reshape(nf, r)
 
    # Af performs the multiplications of exp(ar) by the matrix A, for all frequencies
    # the repeat and transpose functions are tricks to make possible the vector calculation 
    Af = (A.reshape(n, n, 1, r).repeat(nf, axis=2) \
          * np.exp(expoents)).transpose([2, 0, 1, 3])
    # fft sums the value for all lags
    AL = np.eye(n) - np.sum(Af, axis=3)

    return AL


def vech(a):
    ''' Returns vech(v) '''
    n = a.shape[0]
    v = np.empty((n * (n + 1)) / 2)
    cont = 0
    for j in np.arange(n):
        for i in np.arange(n):
            if i >= j:
                v[cont] = a[i, j]
                cont = cont + 1
    return v


def Dup(n):
    '''D*vech(A) = vec(A), with symmetric A'''
    d = np.zeros([n * n, round((n * (n + 1)) / 2)])
    count = 0
    for j in np.arange(n):
        for i in np.arange(n):
            if i >= j:
                d[j * n + i, count] = 1
                count = count + 1
            else:
                d[j * n + i, :] = d[i * n + j, :]
    return d


def TT(a, b):
    ''' TT(a,b)*vec(B) = vec(B.T), where B is (a x b).'''
    t = O(a * b)
    for i in np.arange(a):
        for j in np.arange(b):
            t[i * b + j, j * a + i] = 1
    return coo_matrix(t, dtype = np.int8)


def fdebig_de(n):
    '''Derivative of kron(I(2n), A) by A'''
    a = sparse.kron(TT(2 * n, n), I(n * 2 * n))
    b = sparse.kron(vec(I(2 * n)), I(n))
    #c = coo_matrix.transpose(sparse.kron(I(n), b ))
    c = sparse.kron(I(n), b )
    return  coo_matrix(a.dot(c), dtype = int8)


def fdebig_de_small(n):
    '''Derivative of kron(I(2), A) by A'''
    return dot(sparse.kron(TT(2, n), I(n * 2)),
               sparse.kron(I(n), sparse.kron(vec(I(2)), I(n))))


def xlag(x, lag):
    if(lag == 0):
        return x.copy()
    xl = np.zeros(x.shape)
    xl[:, lag:] = x[:, :-lag]
    return xl


# def bigautocorr_old(x, p):
#     '''Autocorrelation. Data in rows. From order 0 to p-1.
#     Output: nxn blocks of autocorr of lags i. (Nuttall Strand matrix)'''
#     y = x[:]
#     nd = x.shape[1]
#     for i in arange(1, p):
#         y = concatenate((y, xlag(x, i)), axis=0)
#     return dot(y, y.T) / nd
#     # return cov(y.T)


# Older version
def bigautocorr(x, p):
    '''Autocorrelation. Data in rows. From order 0 to p-1.
    Output: nxn blocks of autocorr of lags i. (Nuttall Strand matrix)'''
    n, nd = x.shape
    gamma = np.zeros([n * p, n * p])
    for i in arange(p):
        for j in arange(p):
            gamma[i * n:i * n + n, j * n:j * n + n] = dot(xlag(x, i), xlag(x, j).T) / nd
    return gamma

# New version
# def bigautocorr(x, p):
#     '''Autocorrelation. Data in rows. From order 0 to p-1.
#     Output: nxn blocks of autocorr of lags i. (Nuttall Strand matrix)'''
#     n, nd = x.shape
#     gamma = np.zeros([n * p, n * p])
#     for i in arange(p):
#         g = dot(x, xlag(x, i).T) / (nd - i)
#         for j in arange(p - i):
#             gamma[(i + j) * n:(i + j) * n + n, j * n:j * n + n] = g.T
#             gamma[j * n:j * n + n, (i + j) * n:(i + j) * n + n] = g
#     return gamma


def fdh_da(Af, n):
    '''Derivative of vec(H) by vec(A), with H = A^-1 and complex A.'''
    ha = Af.I
    h = -kron(ha.T, ha)

    h1 = cat(h.real, -h.imag, 1)
    h2 = cat(h.imag, h.real, 1)
    hh = cat(h1, h2, 0)
    return hh


def fIij(i, j, n):
    '''Returns Iij of the formula'''
    Iij = np.zeros(n**2)
    Iij[n * j + i] = 1
    Iij = diag(Iij)
    return np.array(kron(I(2), Iij))
    #return csr_matrix(np.kron(I(2), Iij))   # ### sparse


def fIj(j, n):
    '''Returns Ij of the formula'''
    Ij = np.zeros(n)
    Ij[j] = 1
    Ij = np.diag(Ij)
    Ij = np.kron(Ij, I(n))
    return np.kron(I(2), Ij)


def fIi(i, n):
    '''Returns Ii of the formula'''
    Ii = np.zeros(n)
    Ii[i] = 1
    Ii = np.diag(Ii)
    Ii = np.kron(I(n), Ii)
    return np.kron(I(2), Ii)


def fCa(f, p, n):
    '''Returns C* of the formula'''
    C1 = cos(-2 * pi * f * np.arange(1, p + 1))
    S1 = sin(-2 * pi * f * np.arange(1, p + 1))
    C2 = cat(C1.reshape(1, -1), S1.reshape(1, -1), 0)
    return np.kron(C2, identity(n**2))


def fChol(omega):
    # Try Cholesky factorization
    try:
#       L = np.linalg.cholesky(omega, lower=1)
        L = np.linalg.cholesky(omega)
    # If there's a small negative eigenvalue, diagonalize
    except LinAlgError:
        val, vec = eigh(omega)
        print('non-positive eig. in omega:', val[val <= 0])
        L = np.zeros(vec.shape)
        for i in np.arange(len(val)):
            if val[i] < 0.:
                val[i] = 0.
            L[:, i] = vec[:, i] * sqrt(val[i])
        # print 'L', L

    return L


def fEig(L, G2):
    '''Returns the eigenvalues'''
    L = np.array(L)
    G2 = np.array(G2)
    D = L.T @ G2 @ L

    u, s, v = np.linalg.svd(D, full_matrices=True) 
    # s, v = np.linalg.eigh(D) 
    # previous version 
    # d = np.linalg.eigh(D, eigvals_only=True)
    # print("d ", s.shape)
    # the two biggest eigenvalues no matter which values are non negative by
    # construction
    d1 = np.sort(s)
    d = d1[-2:]  # two largest elements
    # d = d[abs(d) > 1E-8]
    if (d.size > 2):
        print('more than two chi-square in the sum:')
        print(d)
    return d


# @jit(nopython=True)
def asymp_pdc(x, A, nf, e_var, p, metric='info', alpha=0.01):
    '''Asymptotic statistics for the three PDC formulations
        x -> data
        A -> autoregressive matrix
        nf -> number of frequency
        e_var -> residues
        p -> autoregressive model order
        metric -> witch PDC (iPDC = 'info', gPDC = 'diag', PDC = 'euc')
        alpha -> confidence margin
    '''
        
    print(A)
    
    A = np.array(A)
    #x = np.array(x)
    #e_var = np.array(e_var)
    Af = A_to_f(A, nf)
    n, nd = x.shape   # n: number of channels; nd: data points.

    pdc2 = np.empty([n, n, nf])
    
    if alpha != 0:
        th = np.empty([n, n, nf])
        ic1 = np.empty([n, n, nf])
        ic2 = np.empty([n, n, nf])
        varass1 = np.empty([n, n, nf])
        varass2 = np.empty([n, n, nf])
        patdfr = np.empty([n, n, nf])
        patdenr = np.empty([n, n, nf])
        pvalues = np.empty([n, n, nf])

    start0 = timer()
    # dpdc_dev = np.zeros(int((n*(n + 1))/2))
    if metric == 'euc':
        dpdc_dev = np.zeros(int((n*(n + 1))/2))
    elif metric == 'diag':
        evar_d = mdiag(e_var)
        evar_d_big = np.kron(I(2*n), evar_d)
        # inv_ed = evar_d_big.I
        pinv_evar_d_big = np.linalg.pinv(evar_d_big)

        #'derivative of vec(Ed-1) by vecE'
        de_deh = Dup(n)
        debig_de = fdebig_de(n)  #sparsed
        tmp = -pinv_evar_d_big @ pinv_evar_d_big
        tmp3 = vec3(tmp)
        #dedinv_dev = diagtom(vec(-inv_ed * inv_ed))
        dedinv_dev = diagtom(tmp3.toarray())
        dedinv_deh = dedinv_dev @ debig_de @ de_deh
    elif metric == 'info':
        evar_d = mdiag(e_var)   # (n, n)
        evar_d_big = np.kron(I(2 * n), evar_d)    # (2n^2, 2n^2)

        #pinv_evar_d_big = coo_matrix(np.linalg.pinv(evar_d_big)) #
        pinv_evar_d_big = np.linalg.pinv(evar_d_big) # (2n^2, 2n^2)

        #inv_ed = evar_d_big
        evar_big = np.kron(I(2 * n), e_var) # (2n^2, 2n^2)
        pinv_evar_big = np.linalg.pinv(evar_big)  #  (2n^2, 2n^2)

        'derivada de vec(Ed-1) por vecE'
        de_deh = Dup(n)  # (n^2, sum(1:n))

        debig_de = fdebig_de(n)  # sparsed (4n^4, n^2)
        tmp = -pinv_evar_d_big @ pinv_evar_d_big

        pinv_evar_big = csr_matrix(pinv_evar_big)
        # tmp2 = vec(tmp).toarray()   # (2n^2, 2n^2)
        tmp3 = vec3(tmp)
        dedinv_devd = diagtom(tmp3.toarray()) #sparse # (4n^4, 4n^4) for n=8, duration = 24.34 sec

        dedinv_dehd = dedinv_devd @ (debig_de @ de_deh) #sparse (4n^4, sum(1:n))

        dedinv_dev = csr_matrix(-sparse.kron(pinv_evar_big.T, pinv_evar_big))  # (4n^4, 4n^4)
 
        dedinv_deh = coo_matrix(dedinv_dev @ debig_de @ de_deh) #sparse
        # (4n^4, sum(1:n)) = (4n^4, 4n^4) @ (4n^4, n^2) @ (n^2, sum(1:n))

    else:
        pass
        #raise SystemExit("Error: metric should be either 'euc', 'diag' or 'info'.")

    duration1 = timer() - start0
    start1 = timer()
    
    gamma = bigautocorr(x, p)
    omega = np.kron(inv(gamma), e_var)
    omega_evar = 2 * np.linalg.pinv(Dup(n)) @ np.kron(e_var, e_var) \
                                            @ np.linalg.pinv(Dup(n)).T

    icdf_norm_alpha = st.norm.ppf(1 - alpha / 2.0)

    duration2 = timer() - start1

    start2 = timer()

    print('alpha = ', alpha)

    # pinv_evar_d_big = csr_matrix(pinv_evar_d_big)
    duration_ff = 0
    for ff in range(nf): # Frequency loop
        start_ff = timer()

        f = ff / (2.0 * nf)
        Ca = fCa(f, p, n)

        a = vec2(Af[ff, :, :])
        a = cat(a.real, a.imag, 0)
        #a = vec(cat(I(n), O(n), 1)) - dot(Ca, al)

        omega2 = Ca @ omega @ Ca.T
        L = fChol(omega2)

        for j in range(n): # Channel loop j (columnwise)
            Ij = csr_matrix(fIj(j, n))

            if metric == 'euc':
                Ije = Ij
            elif metric == 'diag':
                Ije = csr_matrix(Ij @ pinv_evar_d_big)
            elif metric == 'info':
                Ije = csr_matrix(Ij @ pinv_evar_big @ Ij)
            else:
                pass

            for i in range(n):  # Channel loop i (rowwise)

                # t0 = timer()
                Iij = fIij(i, j, n)

                # For diag or info case, add evar in the formula'

                if metric == 'euc':
                    Iije = Iij  # = Iij.toarray()  # if fIij returns csr_matrix
                else:  # metric == 'diag' or 'info'
                    Iije = Iij @ pinv_evar_d_big

                # t0 = timer()
                num = a.T @ Iije @ a
                den = a.T @ Ije @ a
                pdc2[i, j, ff] = num / den

                # 'Add evar derivation'
                if metric == 'diag':
                    #'derivada do num por vecE'
                    dnum_dev = sparse.kron((Iij @ a.T).T, a) @ dedinv_deh
                    #'derivada do den por vecE'
                    dden_dev = sparse.kron((Ij @ a).T, a) @ dedinv_deh
                    dpdc_dev = (den * dnum_dev - num * dden_dev) / (den**2)
                elif metric == 'info':
                    #'derivada do num por vecE'
                    dnum_dev = sparse.kron((Iij @ a.T).T, a) @ dedinv_dehd
                    #'derivada do den por vecE'
                    dden_dev = sparse.kron((Ij @ a.T).T, a @ Ij) @ dedinv_deh
                    dpdc_dev = (den * dnum_dev - num * dden_dev) / (den**2)
                else: # metric == 'euc'
                    pass

                t0 = timer()
                G1a = 2 * a.T @ Iije / den - 2 * num * a.T @ Ije / (den**2)
                G1 = -G1a @ Ca

                varalpha = G1 @ omega @ G1.T
                varevar = dpdc_dev @ omega_evar @ (dpdc_dev.T)
                varass1[i, j, ff] = (varalpha + varevar) / nd

                ic1[i, j, ff] = pdc2[i, j, ff] \
                                - sqrt(varass1[i, j, ff]) * icdf_norm_alpha
                ic2[i, j, ff] = pdc2[i, j, ff] \
                                + sqrt(varass1[i, j, ff]) * icdf_norm_alpha
                
                G2a = Iije / den
                #G2 = Ca.T*G2a*Ca

                d = fEig(L.real, G2a.real)

                patdf = sum(d)**2 / sum(d**2)
                patden = sum(d) / sum(d**2)
                th[i, j, ff] = st.chi2.ppf(1 - alpha, patdf) / (patden * nd)
                pvalues[i, j, ff] = 1 - st.chi2.cdf(pdc2[i, j, ff] * patden * nd, patdf)

                varass2[i, j, ff] = patdf / (patden * nd)**2
                patdfr[i, j, ff] = patdf
                patdenr[i, j, ff] = patden
                # print("Duração 4 = ", n*n*nf*(timer()-t0))
        duration_ff = timer() - start_ff

    duration3 = timer() - start2
    print("Duração 1 = ", duration1)
    print("Duração 2 = ", duration2)
    print("Duração 3 = ", duration3)

    #ss = ss_alg(A, e_var, nf)
    ss = []
    #coh = coh_alg(ss)
    coh2 = []

    return pdc2, ss, coh2, th, ic1, ic2, patdfr, patdenr
          #mes, ss, coh2, th, ic1, ic2, patdf, patden 

def asymp_dtf(x, A, nf, e_var, p, alpha=0.05):
    '''Asymptotic statistics for the DTF
        x -> data
        A -> autoregressive matrix
        nf -> number of frequencies
        e_var -> residues
        p -> model order
        alpha -> confidence margin
    '''

    #x = np.array(x)
    #e_var = np.array(e_var)
    Af = A_to_f(A, nf)

    n, nd = x.shape

    th = np.empty([n, n, nf])
    ic1 = np.empty([n, n, nf])
    ic2 = np.empty([n, n, nf])
    dtf2 = np.empty([n, n, nf])
    varass1 = np.empty([n, n, nf])
    varass2 = np.empty([n, n, nf])

    gammai = inv(bigautocorr(x, p))
    omega = kron(gammai, e_var)

    for ff in range(nf):
        f = ff / (2.0 * nf)
        Ca = fCa(f, p, n)

        #Hf = np.array(Af[ff, :, :]).I
        Hf = Af[ff, :, :].I
        h = vec(Hf)
        h = cat(h.real, h.imag, 0)

        #dhda = fdh_da(np.array(Af[ff, :, :]), n)
        dhda = fdh_da(Af[ff, :, :], n)

        #L = fChol(omega)

        omega2 = dhda * Ca * omega * Ca.T * dhda.T
        L = fChol(omega2)

        for i in range(n):
            for j in range(n):

                Iij = fIij(i, j, n)


                Ii = fIi(i, n)

                num = h.T * Iij * h
                den = h.T * Ii * h
                dtf2[i, j, ff] = num / den

                G1a = 2 * h.T * Iij / den - 2 * num * h.T * Ii / (den**2)
                G1 = -G1a * dhda * Ca

                varass1[i, j, ff] = G1 * omega * G1.T / nd
                ic1[i, j, ff] = dtf2[i, j, ff] - sqrt(varass1[i, j, ff]) * st.norm.ppf(1 - alpha / 2)
                ic2[i, j, ff] = dtf2[i, j, ff] + sqrt(varass1[i, j, ff]) * st.norm.ppf(1 - alpha / 2)

                G2a = 2 * Iij / den
                #G2 = Ca.T*dhda.T*G2a*dhda*Ca
                G2 = G2a

                d = fEig(L, G2)
                patdf = sum(d)**2 / sum(d**2)
                patden = sum(d) / sum(d**2)

                th[i, j, ff] = st.chi2.ppf(1 - alpha, patdf) / (patden * 2 * nd)
                varass2[i, j, ff] = 2 * patdf / (patden * 2 * nd)**2

    return dtf2, th, ic1, ic2, patdf, patden


def fc(i, n):
    vi = np.zeros(n)
    vi[i] = 1
    vi.resize(n, 1)
    return np.kron(vi, I(n))


def fk1(e_var_inv, i, j, n):
    ci = fc(i, n)
    cj = fc(j, n)

    return np.kron(I(2), ci) @ e_var_inv @ np.transpose(np.kron(I(2), cj))


def fk2(e_var_inv, i, j, n):
    ci = fc(i, n)
    cj = fc(j, n)

    return np.kron(I(2), ci) @ e_var_inv @ np.transpose(np.kron(np.array([[0, 1], [-1, 0]]), cj))


def fl(i, n):
    vi = np.zeros(n)
    vi[i] = 1
    vi.resize(n, 1)
    return np.kron(I(n), vi)


def fkl1(evar_big, i, j, n):
    li = fl(i, n)
    lj = fl(j, n)

    return np.kron(I(2), li) @ evar_big @ np.transpose(np.kron(I(2), lj))


def fkl2(evar_big, i, j, n):
    li = fl(i, n)
    lj = fl(j, n)

    return np.kron(I(2), li) @ evar_big @ np.transpose(np.kron(array([[0, 1], [-1, 0]]), lj))


def ss_alg(A, e_cov, nf):
    '''Calculates the Spectral density (SS)
         A -> autoregressive matrix
         e_cov -> residues
         nf -> number of frequencies
    '''
    n, n, r = A.shape
    AL = A_to_f(A, nf)
    ss = np.empty(AL.shape)
    for i in range(n):
        #H = inv(reshape(AL[i,:,:],n,n))
        H = inv(AL[i,:,:].reshape(n, n, order='F').copy())
        ss[i,:,:] = H * e_cov * H.T
 
    #print ss[5]
    #SS=permute(ss,[2,3,1])
    SS = ss[:,:,None].transpose([1, 2, 0]) # or equivalently
    # SS = np.transpose( np.expand_dims(ss, axis=2), (1, 2, 0) )
    return SS   #MATLAB    return ss.transpose(1,2,0)


def coh_alg(SS):
    """Coh2 = coh_alg(SS)"""
    m, n, nf = SS.shape
    Coh2 = np.empty([m, n, nf])
    if m == n: 
        nChannels = m
    else:
        print('Wrong SS dimension.')

    for k in range(nf):
        for iu in range(nChannels):
            for ju in range(nChannels):
                Coh2[iu, ju, k] = SS[iu, ju, k] / sqrt(SS[iu,iu,k] * SS[ju, ju, k])

    return Coh2


def fCij(i, j, n, p):
    '''Returns Cij of the formula'''
    Cij = np.zeros([1, n * n])
    Cij[0, (j * n + i)] = 1
    Cij = np.kron(I(p), Cij)
    return Cij


def asymp_gct(x, A, e_var):
    '''Asymptotic statistics for Wald statistic of the GC in time
        x -> data
        A
        e_var
    '''

    x = np.array(x)
    e_var = np.array(e_var)

    n, nd = x.shape
    n, n, p = A.shape

    wt = np.empty([n, n])

    gammai = inv(bigautocorr(x, p))
    omega = kron(gammai, e_var)

    a = vec(A)

    for i in range(n):
        for j in range(n):
            Cij = fCij(i, j, n, p)
            wt[i, j] = (nd * 1.0 / p) * (Cij * a).T * ((Cij * omega * Cij.T).I) * (Cij * a)

    pv = 1 - st.f.cdf(wt, p, nd - n * p - 1)

    return pv, wt


def asymp_igct(e_var, nd):
    '''Asymptotic statistics for Wald statistic of instantaneous GC
        e_var
        nd
    '''

    e_var = np.array(e_var)

    n, n = e_var.shape

    wt = np.zeros([n, n])

    for i in np.arange(n):
        for j in np.arange(n):
            Cij = np.zeros([n, n])
            Cij[i, j] = 1
            Cij[j, i] = 1
            Cij = np.array(vech(Cij))
            eij = e_var[i, j]
            Di = Dup(n).I
            wt[i, j] = nd * eij * (2 * Cij * Di * kron(e_var, e_var) * Di.T * Cij.T).I * eij

    pv = 1 - st.chi2.cdf(wt, 1)

    return pv, wt


def asymp_white(x, res, p, h=20):

    n, nd = res.shape

    x = np.empty([n, n, h + 1])

    for i in np.arange(n):
        for j in np.arange(n):
            dum, aux, dum, dum = xcorr(res[i], res[j], maxlags=h, normed=False)
            x[i, j] = aux[h:]

    s0 = np.array(inv(x[:, :, 0]))
    s = 0
    for i in arange(1, h + 1):
        s = s + (1.0 / (nd - i)) * trace(x[:, :, i].T * s0 * x[:, :, i] * s0)
    s = s * nd**2
    pv = 1 - st.chi2.cdf(s, (h - p) * n**2)

    return pv, s



from glob import glob
import os
import sys
from setuptools import setup


name = "pypdc"
version="0.0.6"
description = "Python asymptotic Partial Directed Coherence and Directed Coherence estimation package for brain connectivity analysis."
authors = {
    "Sameshima": ("Koichi Sameshima", "ksameshi@usp.br"),
    "Brito": ("Carlos Stein Naves de Brito", "c.brito@ucl.ac.uk"),
    "Baldo" : ("Heitor Baldo", "hbaldo@usp.br")
}

platforms = ["Linux", "Mac OSX", "Windows", "Unix"]
keywords = [
    "Brain Connectivity",
    "PDC", "iPDC", 
    "Granger Causality",
    ]
classifiers = [
    "Development Status :: 5 - Production/Stable",
    "Intended Audience :: Developers",
    "Intended Audience :: Science/Research",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: 3 :: Only",
    "Topic :: Software Development :: Libraries :: Python Modules",
    "Topic :: Scientific/Engineering :: Bio-Informatics",
]


packages = ["pypdc"]
    

with open("README.rst", "r") as fh:
    long_description = fh.read()

if __name__ == "__main__":

    setup(
        name=name,
        version=version,
        author=authors["Sameshima"][0],
        author_email=authors["Sameshima"][1],
        description=description,
        keywords=keywords,
        platforms=platforms,
        classifiers=classifiers,
        packages=packages,
        zip_safe=False,
    )


'''
This code plots the connectivity digraph such that the node positions are  
in 10-20 system style configuration.
'''

import numpy as np
import networkx as nx
import matplotlib.pyplot as plt


#fixed positions:
pos = {0: (0,15), 1: (2,16), 2: (4,16), 3: (6, 15), 
       4: (-2,13), 5: (0,12), 6: (2,12), 7: (4,12), 8: (6,12), 9: (8, 13),
       10: (-2,8), 11: (0, 8), 12: (2, 8), 13: (4, 8), 14: (6, 8), 15: (8, 8),
       16: (-2,3), 17: (0, 4), 18: (2, 4), 19: (4, 4), 20: (6, 4), 21: (8, 3),
       22: (2, 0), 23: (4, 0)}

#Setting a 24x24 matrix (24 channels):
def set_matrix(M):
    A = np.zeros((24,24))
    arr = np.array([i for i in range(len(M))])
    A[arr[:, None], arr] += M
    return A

#Plotting the connectivity digraph:
def plot_digraph(M):
    A = set_matrix(M)
    G = nx.from_numpy_matrix(A, create_using=nx.MultiDiGraph())

    plt.figure(3,figsize=(8,8)) 
    nx.draw(G, pos=pos, with_labels=True, node_color="#ffffff", node_size=580,
            connectionstyle='arc3, rad = 0.1', edgecolors='black', font_weight='bold') #, width=weights)
    plt.show()
